unit U_PosPosExTest;
{Copyright  2007, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  {$IF declared(CompilerVersion) and (CompilerVersion>=15.0)}
    {$define TestPosEx}
    strutils,
  {$IFEND}
  {For Delphi 7 & later  "PosEx" test}
  StdCtrls, ShellAPI;

type
  TForm1 = class(TForm)
    PostestBtn: TButton;
    Edit2: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    PosExBtn: TButton;
    Memo1: TMemo;
    Memo2: TMemo;
    StaticText1: TStaticText;
    Memo3: TMemo;
    procedure PostestBtnClick(Sender: TObject);
    procedure PosExBtnClick(Sender: TObject);
    procedure StaticText1Click(Sender: TObject);
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

{*************** Pos ****************}
function pos(const substr, str:string):integer;
{Search for "substr" in "str". Return 0 or the starting postion in "str" where
 "substr" was found}
var
  i,j,k,ssLen, sLen, stop:integer;
  a:char;
begin
  result:=0;
  ssLen:=length(substr);
  slen:=length(str);
  stop:=slen-sslen+1; {highest start location for substring}
  if (ssLen=0) or (sslen>sLen) then exit;
  a:=substr[1];  {1st letter of substr}
  i:=1; {start search location}
  while (i<=stop) and (result=0) do
  begin
    while (i<=stop) and (a<>str[i]) do inc(i);
    if i<=stop then
    begin
      if sslen=1 then  result:=i
      else
      begin
        j:=2;
        k:=i+1;
        while (j<=sslen) and (k<=slen) do
        begin
          if (substr[j]<>str[k]) then break;
          inc(j);
          inc(k);
        end;
        if (j>sslen) {and (substr[j]=str[k])} then  result:=i
        else inc(i);
      end;
    end;
  end;
end;



{*********** PosEx ***********}
function Posex(const substr,str:string; const startloc:integer):integer;
{Search for "substr" in "str" starting at "startloc" return 0 or the start
 postion where "substr" was found}
var
  i, j,k,ssLen, sLen, stop:integer;
  a:char;
begin
  result:=0;
  ssLen:=length(substr);
  slen:=length(str);
  stop:=slen-sslen+1; {highest feasible start location for substring}
  if (ssLen=0) or (sslen>sLen) then exit;
  a:=substr[1];  {1st letter of substr}
  i:=startloc; {start search location}
  while (i<=stop) and (result=0) do
  begin
    while (i<=stop) and (a<>str[i]) do inc(i); {find the 1st letter}
    if i<=stop then
    begin
      if sslen=1 then  result:=i {It was a 1 character search, so we're done}
      else
      begin
        j:=2;
        k:=i-1; {back "K" up by 1 so that we can use K+j as the index to the string}
        while (j<=sslen) do
        begin {compare the rest of the substring}
          if (substr[j]<>str[k+j]) then break
          else inc(j); {The letter matched, go to the next+
                   {When we pass the substring end, "while" loop will terminate}
        end;
        if (j>sslen) then
        begin
          result:=i;
          exit;
        end
        else inc(i); {that search failed, look for the next 1st letter match}
      end;
    end;
  end;
end;



{************* PosTestBtnClick ***********}
procedure TForm1.PostestBtnClick(Sender: TObject);
var
  i,n1,n2:integer;
  freq,start,stop:int64;
  time1,time2:extended;
  s1,s2:string;
begin
  s1:=memo3.text;
  s2:=edit2.text;
  memo1.lines.add(format('Testing "Pos"function searching for "%s" in :',[s2]));
  memo1.Lines.Add(s1);
  N1:=System.pos(s2,s1);
  N2:=pos(s2, s1);
  if N1=N2 then
  begin
    memo1.lines.add('Match found at position '+inttostr(N1));
    queryperformancefrequency(freq);

    queryperformancecounter(start);
    for i:=1 to 1000000 do System.pos(s2,s1);
    queryperformancecounter(stop);
    time1:=(stop-start)/freq;

    queryperformancecounter(start);
    for i:=1 to 1000000 do pos(s2,s1);
    queryperformancecounter(stop);
    time2:=(stop-start)/freq;
    Memo1.lines.add(format('1,000,000 calls: System "Pos" time = %6.1f microsecs per call',
              [time1]));
    Memo1.lines.add(format('1,000,000 calls: DFF Test "Pos" time = %6.1f microsecs per call',
              [time2]));
  end
  else memo1.lines.add(format('Error; N(Real)=%d, N(Test)=%d',[N1,n2]));
  memo1.lines.add('');
end;

(*
{Delphi 7 version of Posex}
function PosEx(const SubStr, S: string; Offset: Cardinal = 1): Integer;
var
  I,X: Integer;
  Len, LenSubStr: Integer;
begin
  if Offset = 1 then
    Result := Pos(SubStr, S)
  else
  begin
    I := Offset;
    LenSubStr := Length(SubStr);
    Len := Length(S) - LenSubStr + 1;
    while I <= Len do
    begin
      if S[I] = SubStr[1] then
      begin
        X := 1;
        while (X < LenSubStr) and (S[I + X] = SubStr[X + 1]) do
          Inc(X);
        if (X = LenSubStr) then
        begin
          Result := I;
          exit;
        end;
      end;
      Inc(I);
    end;
    Result := 0;
  end;
end;
*)

{************ PosExtestBtnClick **********}
procedure TForm1.PosExBtnClick(Sender: TObject);
{Find all occurrences of a substring in a string}

var
  i,j,n1,n2:integer;
  freq,start,stop:int64;
  time:extended;
  s1,s2:string;
begin

  s1:=memo3.text;  {the string}
  s2:=edit2.text;  {the substring}
  memo1.lines.add(format('Testing "PosEx"function searching for all occurrences of "%s" in:',[s2]));
  memo1.lines.add(format('%s',[s1]));
  i:=1;
  N1:=0;
  N2:=0;
  while (i<length(s1)-length(s2))do
  begin
    n2:=posex(s2,s1,i);

    {$IfDEF TestPosEx}
      n1:=StrUtils.posex(s2,s1,i);
    {$Else}
      N1:=n2;
    {$EndIf}
    if n1<>n2 then break;

    if N2>0 then
    begin
      memo1.lines.add('Match found at position '+inttostr(N2));
      i:=N2+1;
    end
    else break;
  end;
  if N1=N2 then
  begin
    queryperformancefrequency(freq);

    {$IfDEF TestPosEx}
      queryperformancecounter(start);
      for j:=1 to 1000000 do
      begin
        i:=1;
        while i<=length(s1)-length(s2) do
        begin
          n1:=strutils.posex(s2,s1,i);
          if n1>0 then i:=n1+1 else break;
        end;
      end;
      queryperformancecounter(stop);
      time:=(stop-start)/freq;
      memo1.lines.add(format('1,000,000, calls:  System "PosEx" time = %6.1f microsecs per call, ',
                  [time]));
    {$EndIf}

    {Test DFF version}
    queryperformancecounter(start);
    for j:=1 to 1000000 do
    begin
      i:=1;
      while i<=length(s1)-length(s2) do
      begin
        n2:=posex(s2,s1,i);
        if n2>0 then i:=n2+1 else break;
      end;
    end;
    queryperformancecounter(stop);
    time:=(stop-start)/freq;
    memo1.lines.add(format('1,000,000 calls: DFF Test "PosEx" time = %6.1f microsecs per call, ',
                  [time]));
  end
  else memo1.lines.add(format('PosEx Error: N(Delphi)=%d, N(Test)=%d',[N1,n2]));
  memo1.lines.add('');
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
  ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;


end.
